#include "stdafx.h"
#include "BloomFilterOptimisation.h"
#include <climits>
#include <crtdbg.h>
#include "InstanceData.h"
#include "StatisticsCollection.h"

void ProduceBfBits()
{
	int *resourceMap = new int[numberOfResources];
	int numberOfBits = 0;
	for (int resource = 0; resource < numberOfResources; resource++)
	{
		resourceMap[resource] = numberOfBits;
		numberOfBits += resourceCapacities[resource] + 1;
	}


	// Extra bit for the "closing" element in the list
	bfBits = new BloomFilterBit[numberOfBits + 1];

	int curIndex = 0;
	for (int resource = 0; resource < numberOfResources; resource++)
		for (int capacity = 0; capacity <= resourceCapacities[resource]; capacity++)
		{
			bfBits[curIndex].resource = resource;
			bfBits[curIndex].minVolume = capacity;
			bfBits[curIndex].availabilityProbability = availabilityStat[resource][capacity];
			bfBits[curIndex].demandProbability = 0; // Resetting; calculation is below
			bfBits[curIndex].nextIndex = curIndex + 1;
			curIndex++;
		}

	// "Closing" element
	bfBits[curIndex].resource = -1;

	for (JobInfo *jobInfo = jobInfos + numberOfJobs - 1; jobInfo >= jobInfos; jobInfo--)
		for (int *resource = jobInfo->consumedResources, *volume = jobInfo->consumedVolumes;
			*resource >= 0;
			resource++, volume++)
		{
			bfBits[resourceMap[*resource] + *volume].demandProbability++;
		}

	delete[] resourceMap;
}

void OptimiseBfHeuristic()
{
	int numberOfBits = 0;
	for (BloomFilterBit *curBit = bfBits; curBit->resource >= 0; curBit = bfBits + curBit->nextIndex)
		numberOfBits++;

	// While there are too many bits
	while (numberOfBits > sizeof(int) * 8)
	{
		int bestAvailIndex = -1;
		int minProb = INT_MAX;

		// Find the pair of bits to merge
		for (int availIndex = 0; bfBits[availIndex].resource >= 0; availIndex = bfBits[availIndex].nextIndex)
		{
			BloomFilterBit *availBit = bfBits + availIndex;
			BloomFilterBit *demandBit = bfBits + availBit->nextIndex;

			if (availBit->resource == demandBit->resource)
			{
				int failProb = demandBit->demandProbability * availBit->availabilityProbability;
				if (failProb < minProb)
				{
					minProb = failProb;
					bestAvailIndex = availIndex;
				}
			}
		}

		// Merge selected bits
		{
			BloomFilterBit *availBit = bfBits + bestAvailIndex;
			BloomFilterBit *demandBit = bfBits + availBit->nextIndex;

			// Update probabilities
			availBit->availabilityProbability += demandBit->availabilityProbability;
			availBit->demandProbability += demandBit->demandProbability;

			// Remove demandBit from the list
			availBit->nextIndex = demandBit->nextIndex;

			numberOfBits--;
		}
	}
}


void ProduceBfDictionary()
{
	bfDictionary = new int*[numberOfResources];

	int shift = 0;
	BloomFilterBit *curBit = bfBits;
	for (int resource = 0; resource < numberOfResources; resource++)
	{
		int mapSize = resourceCapacities[resource] + 1;
		int *dict = bfDictionary[resource] = new int[mapSize];

		_ASSERT(!curBit->minVolume);

		int usedBits = -1;
		dict[0] = 0;
		for (int volume = 0; volume < mapSize; volume++)
		{
			if (curBit->resource == resource && curBit->minVolume == volume)
			{
				usedBits++;
				curBit = bfBits + curBit->nextIndex;
			}

			dict[volume] = ((1 << usedBits) - 1) << shift;
		}

		shift += usedBits;
	}
}