// NativeImplementation.cpp : Defines the exported functions for the DLL application.

//#define STORE_STATS_IN_FILE
//#define STORE_RUNTIMES


#include "stdafx.h"
#include "BloomFilterOptimisation.h"
#include "NoBloomFilters.h"
#include "WithBloomFilters.h"
#include <cstring>
#include "StatisticsCollection.h"
#include <cstdio>
#include <cstdlib>
#include "AlgorithmParameters.h"
#include "SignTest.h"

int executionsSinceStatsCollection;
int statsCollectionsUntilUpdate;
int horizonUpperBound;


enum {
	MODE_HYBRID = 0,
	MODE_NONBF = 1,
	MODE_BF = 2,
	MODE_CONVENTIONAL = 3
};

int mode;

extern "C"
{
	__declspec(dllexport) void InitScheduleGenerator(
		int numberOfJobs, 
		int numberOfResources, 
		int *durations, 
		int *consumption, 
		int *resourceCapacities, 
		int *predecessors,
		int *successors,
		int mode);

	__declspec(dllexport) int ScheduleGenerator(int *solution, int *times);

	__declspec(dllexport) int Metaheuristic(
		int *curSolution,
		int *bestSolution,
		int *times,
		int iterations,
		int randomSeed,
		int *runtimes);

	__declspec(dllexport) int GetFalsePositives();
}


__declspec(dllexport) int GetFalsePositives()
{
	return falsePositives;
}

void ClearMemory()
{
	delete[] ::resourceCapacities;

	for (int resource = 0; resource < ::numberOfResources; resource++)
		delete[] resourceAvailability[resource];

	delete[] resourceAvailability;

	delete[] jobInfos;
	delete[] consumptionFlat;
	delete[] predecessorsFlat;
	delete[] successorsFlat;
	delete[] consumedResourcesFlat;
	delete[] consumedCapacitiesFlat;

	delete[] findNonBfFunctions;
	delete[] findStatFunctions;
	delete[] updateNonBfFunctions;

	if (bloomFilters)
	{
		delete[] bloomFilters;
		bloomFilters = nullptr;

		for (int resource = 0; resource < ::numberOfResources; resource++)
			delete[] bfDictionary[resource];

		delete[] bfDictionary;
		delete[] bfResourcesFlat;
		delete[] bfCapacitiesFlat;
		delete[] findBfFunctions;
		delete[] updateBfFunctions;
	}

	if (mode != MODE_CONVENTIONAL)
		ClearStatistics();
}


void InitialiseBloomFiltersArray()
{
	bfFullAvailability = 0;
	for (int resource = 0; resource < numberOfResources; resource++)
		bfFullAvailability |= bfDictionary[resource][resourceCapacities[resource]];

	bloomFilters = new int[horizonUpperBound + 1];
	for (int slot = 0; slot <= horizonUpperBound; slot++)
		bloomFilters[slot] = bfFullAvailability;
}

int ComputeHorizonUpperBound()
{
	int upperBound = 0;
	for (int job = 0; job < numberOfJobs; job++)
		upperBound += jobInfos[job].duration;

	return upperBound;
}

// Should be called every time bfDictionary changes
void UpdateBfAuxiliaryData()
{
	int *unusedBfResourcesFlat = bfResourcesFlat;
	int *unusedBfCapacitiesFlat = bfCapacitiesFlat;

	for (int job = numberOfJobs - 1; job >= 0; job--)
	{
		JobInfo *jobInfo = jobInfos + job;
		int numberOfConsumedResources = 0;
		while (jobInfo->consumedResources[numberOfConsumedResources] >= 0)
			numberOfConsumedResources++;

		jobInfo->resourcesToTest = unusedBfResourcesFlat;
		jobInfo->volumesToTest = unusedBfCapacitiesFlat;

		int numberOfUncertainBfResources = 0;
		for (int *resource = jobInfo->consumedResources, *volume = jobInfo->consumedVolumes; 
			*resource >= 0; 
			resource++, volume++)
		{
			int* curDictionary = bfDictionary[*resource];
			if (curDictionary[*volume] == curDictionary[*volume - 1])
			{
				jobInfo->resourcesToTest[numberOfUncertainBfResources] = *resource;
				jobInfo->volumesToTest[numberOfUncertainBfResources] = *volume;
				numberOfUncertainBfResources++;
			}
		}
		jobInfo->resourcesToTest[numberOfUncertainBfResources] = -1;
		jobInfo->volumesToTest[numberOfUncertainBfResources] = -1;
		unusedBfResourcesFlat += numberOfUncertainBfResources + 1;
		unusedBfCapacitiesFlat += numberOfUncertainBfResources + 1;
		_ASSERT(numberOfUncertainBfResources <= numberOfResources);

		if (numberOfConsumedResources == 1)
		{
			findBfFunctions[job] = FindEarliestSlotTestSingleResource;
			updateBfFunctions[job] = UpdateResourseAvailabilityBFSingle;
			
			//jobInfo->findEarliestSlot = FindEarliestSlotTestSingleResource;
			//jobInfo->findEarliestSlotCollectStats = FindEarliestSlotTestSingleResource;
			//jobInfo->updateResourceAvailability = UpdateResourseAvailabilityBFSingle;
		}
		else
		{
			//jobInfo->findEarliestSlotCollectStats = FindEarliestSlotCollectStats;

			switch (numberOfUncertainBfResources)
			{
			case 0:
				findBfFunctions[job] = FindEarliestSlotTestBfOnly;
				break;

			case 1:
				findBfFunctions[job] = FindEarliestSlotTestBfAndSingleResource;
				break;

			default:
				findBfFunctions[job] = FindEarliestSlotTestBfAndSomeResources;
				break;
			}
			if (!numberOfUncertainBfResources)
				findBfFunctions[job] = FindEarliestSlotTestBfOnly;
			else if (numberOfUncertainBfResources == 1)
				findBfFunctions[job] = FindEarliestSlotTestBfAndSingleResource;
			else
				findBfFunctions[job] = FindEarliestSlotTestBfAndSomeResources;

			updateBfFunctions[job] = UpdateResourceAvailabilityBf;
		}

		jobInfo->unconsumedResourcesBfMask = -1;
		jobInfo->bloomFilter = 0;
		for (int *resource = jobInfo->consumedResources, *volume = jobInfo->consumedVolumes;
			*resource >= 0;
			resource++, volume++)
		{
			jobInfo->bloomFilter |= bfDictionary[*resource][*volume];
			jobInfo->unconsumedResourcesBfMask &= ~bfDictionary[*resource][resourceCapacities[*resource]];
		}
	}
}

__declspec(dllexport) void InitScheduleGenerator(
	int numberOfJobs, 
	int numberOfResources, 
	int *durations1, 
	int *consumption, 
	int *resourceCapacities, 
	int *predecessors, 
	int *successors,
	int mode)
{
	if (::numberOfJobs)
		ClearMemory();
	
	::mode = mode;

	// Save instance
	::numberOfJobs = numberOfJobs;
	::numberOfResources = numberOfResources;

	::resourceCapacities = new int[numberOfResources];
	memcpy(::resourceCapacities, resourceCapacities, numberOfResources * sizeof(int));

	consumptionFlat = new int[numberOfJobs * numberOfResources];
	memcpy(consumptionFlat, consumption, numberOfResources * numberOfJobs * sizeof(int));

	int totalPredecessorsArraySize = 0;
	{
		int job = 0;
		while (job < numberOfJobs)
			if (predecessors[totalPredecessorsArraySize++] < 0)
				job++;
	}

#ifdef _DEBUG
	int totalSuccessorsArraySize = 0;
	{
		int job = 0;
		while (job < numberOfJobs)
			if (successors[totalSuccessorsArraySize++] < 0)
				job++;
	}
#endif


	predecessorsFlat = new int[totalPredecessorsArraySize];
	memcpy(predecessorsFlat, predecessors, totalPredecessorsArraySize * sizeof(int));
	successorsFlat = new int[totalPredecessorsArraySize];
	memcpy(successorsFlat, successors, totalPredecessorsArraySize * sizeof(int));


	// Produce job infos
	consumedResourcesFlat = new int[numberOfJobs * (numberOfResources + 1)];
	consumedCapacitiesFlat = new int[numberOfJobs * (numberOfResources + 1)];

	jobInfos = new JobInfo[numberOfJobs];
	int *curPredecessorList = predecessorsFlat;
	int *curSuccessorsList = successorsFlat;
	for (int job = 0; job < numberOfJobs; job++)
	{
		JobInfo *jobInfo = jobInfos + job;
		jobInfo->duration = durations1[job];

		jobInfo->predecessors = curPredecessorList;
		jobInfo->successors = curSuccessorsList;

		while (*curPredecessorList >= 0)
			curPredecessorList++; 
		curPredecessorList++;

		while (*curSuccessorsList >= 0)
			curSuccessorsList++;
		curSuccessorsList++;

		jobInfo->consumedResources = consumedResourcesFlat + (numberOfResources + 1) * job;
		jobInfo->consumedVolumes = consumedCapacitiesFlat + (numberOfResources + 1) * job;
		int numberOfConsumedResources = 0;

		for (int resource = 0; resource < numberOfResources; resource++)
		{
			int curConsumption = consumption[job * numberOfResources + resource];
			if (!curConsumption && mode != MODE_CONVENTIONAL)
				continue;

			jobInfo->consumedResources[numberOfConsumedResources] = resource;
			jobInfo->consumedVolumes[numberOfConsumedResources] = curConsumption;
			numberOfConsumedResources++;
		}

		jobInfo->consumedResources[numberOfConsumedResources] = -1;
		jobInfo->consumedVolumes[numberOfConsumedResources] = -1;
	}


	// Allocate memory for resource availability
	horizonUpperBound = ComputeHorizonUpperBound();

	resourceAvailability = new int*[numberOfResources];
	for (int resource = 0; resource < numberOfResources; resource++)
	{
		resourceAvailability[resource] = new int[horizonUpperBound + 1];
		for (int slot = 0; slot <= horizonUpperBound; slot++)
			resourceAvailability[resource][slot] = resourceCapacities[resource];
	}

	if (mode != MODE_CONVENTIONAL)
		InitStatistics();


	// Initialise Bloom filters
	switch (mode)
	{
	case MODE_BF:
	case MODE_HYBRID:
		ProduceBfBits();
		OptimiseBfHeuristic();
		ProduceBfDictionary();

		bfResourcesFlat = new int[numberOfJobs * (numberOfResources + 1)];
		bfCapacitiesFlat = new int[numberOfJobs * (numberOfResources + 1)];

		//InitialiseBloomFiltersArray();

		findBfFunctions = new find[numberOfJobs];
		updateBfFunctions = new update[numberOfJobs];

		//UpdateBfAuxiliaryData();
		break;
	}

	findNonBfFunctions = new find[numberOfJobs];
	findStatFunctions = new find[numberOfJobs];
	updateNonBfFunctions = new update[numberOfJobs];

	// Assignment of dynamic functions
	if (mode == MODE_CONVENTIONAL)
		for (int job = numberOfJobs - 1; job >= 0; job--)
		{
			findNonBfFunctions[job] = FindEarliestSlotNaive;
			findStatFunctions[job] = nullptr; // Should not be used in Conventional mode
			updateNonBfFunctions[job] = UpdateResourseAvailabilityMany;
		}
	else
		for (int job = numberOfJobs - 1; job >= 0; job--)
		{

			JobInfo *jobInfo = jobInfos + job;

			int numberOfConsumedResources = 0;
			while (jobInfo->consumedResources[numberOfConsumedResources] >= 0)
				numberOfConsumedResources++;


			if (numberOfConsumedResources == 1)
			{
				findNonBfFunctions[job] = FindEarliestSlotTestSingleResource;
				findStatFunctions[job] = FindEarliestSlotTestSingleResource;
				updateNonBfFunctions[job] = UpdateResourseAvailabilitySingle;
			}
			else if (numberOfResources - numberOfConsumedResources >= 2)
			{
				findNonBfFunctions[job] = FindEarliestSlotTestManyResources;
				findStatFunctions[job] = FindEarliestSlotCollectStats;
				updateNonBfFunctions[job] = UpdateResourseAvailabilityMany;
			}
			else
			{
				findNonBfFunctions[job] = FindEarliestSlotTestManyResources;
				findStatFunctions[job] = FindEarliestSlotCollectStats;
				updateNonBfFunctions[job] = UpdateResourseAvailabilityMany;
			}
		}

	executionsSinceStatsCollection = ExecutionsBetweenRestarts; // To force stats collectin in first iteration
	falsePositives = 0;
}



int GetEarliestStartTime(JobInfo* jobInfo, int* times)
{
	int startTime = 0;
	for (int *predecessor = jobInfo->predecessors; *predecessor >= 0; predecessor++)
	{
		_ASSERT(times[*predecessor] >= 0); // In DEBUG mode, times are initialised with -1's; hence this tests that the predecessor is already scheduled.
		int predecessorEndTime = times[*predecessor] + jobInfos[*predecessor].duration;
		if (predecessorEndTime > startTime)
			startTime = predecessorEndTime;
	}

	return startTime;
}

void RestoreResourceAvailability(int horizon, bool useBloomFilters)
{
	// Restore resource availability for next run of the generator
	for (int resource = numberOfResources - 1; resource >= 0; resource--)
	{
		int *curResourceAvailability = resourceAvailability[resource];
		int curCapacity = resourceCapacities[resource];
		for (int slot = horizon; slot >= 0; slot--)
			curResourceAvailability[slot] = curCapacity;
	}

	// Restore bloom filters for next run of the generator
	if (useBloomFilters)
		for (int slot = horizon; slot >= 0; slot--)
			bloomFilters[slot] = bfFullAvailability;
}



__declspec(dllexport) int ScheduleGenerator(int *solution, int *times, bool useBloomFilters)
{
	int horizon;

#ifdef _DEBUG
	for (int job = 0; job < numberOfJobs; job++)
		times[job] = -1;
#endif

	horizon = 0;
	for (int jobIndex = 0; jobIndex < numberOfJobs; jobIndex++)
	{
		int job = solution[jobIndex];
		JobInfo *jobInfo = jobInfos + job;

		// Identify the earliest time when the job can be scheduled due to precedence relationships
		int startTime = GetEarliestStartTime(jobInfo, times);


		// Find the earliest slot when the job can be scheduled due to the resource availability
		times[job] = startTime = findFunctions[job](jobInfo, startTime);

#ifdef _DEBUG
		for (int curSlot = startTime; curSlot < startTime + jobInfo->duration; curSlot++)
			for (int *resource = jobInfo->consumedResources, *volume = jobInfo->consumedVolumes; *resource >= 0; resource++, volume++)
				_ASSERT(resourceAvailability[*resource][curSlot] >= *volume);
#endif


		// Update resource availability
		updateFunctions[job](jobInfo, startTime);


#ifdef _DEBUG
		for (int curSlot = startTime; curSlot < startTime + jobInfo->duration; curSlot++)
			for (int resource = numberOfResources - 1; resource >= 0; resource--)
				_ASSERT(resourceAvailability[resource][curSlot] >= 0);

		if (useBloomFilters)
			for (int curSlot = startTime; curSlot < startTime + jobInfo->duration; curSlot++)
				_ASSERT(bloomFilters[curSlot] == GetCorrectBf(curSlot));
#endif


		// Update horizon
		int endTime = startTime + jobInfo->duration - 1;
		if (endTime > horizon)
			horizon = endTime;
	}

	// Restore resource availability for next run of the generator
	RestoreResourceAvailability(horizon, useBloomFilters);

	return horizon;
}



__declspec(dllexport) int ScheduleGenerator(int *solution, int *times)
{
	static int bfBetterCounter = 0;
	static int bfTiming;
	static bool tuning;
	static bool bestIsBf;

	// In conventional mode, ConventionalScheduleGenerator should be used.
	_ASSERT(mode != MODE_CONVENTIONAL);

	int horizon;
	if (executionsSinceStatsCollection++ == ExecutionsBetweenRestarts)
	{
		ResetStatistics();

		findFunctions = findStatFunctions;
		updateFunctions = updateNonBfFunctions;
		horizon = ScheduleGenerator(solution, times, false);


		//printf("BF update\n");

#ifdef STORE_STATS_IN_FILE

		FILE *f = fopen("stats.txt", "a");
		if (f)
		{
			for (int resource = 0; resource < numberOfResources; resource++)
			{
				for (int volume = 0; volume <= resourceCapacities[resource]; volume++)
					fprintf(f, "%d\t", availabilityStat[resource][volume]);

				fprintf(f, "\t");
			}
			fprintf(f, "\n");
			fclose(f);
		}
#endif

		switch (mode)
		{
		case MODE_BF:
		case MODE_HYBRID:
			ProduceBfBits();
			OptimiseBfHeuristic();
			ProduceBfDictionary();

			InitialiseBloomFiltersArray();
			UpdateBfAuxiliaryData();
			break;
		}

		OptimiseResourcesOrder();

		switch (mode)
		{
		case MODE_BF:
			findFunctions = findBfFunctions;
			updateFunctions = updateBfFunctions;
			tuning = false;
			bestIsBf = true;
			break;

		case MODE_NONBF:
			findFunctions = findNonBfFunctions;
			updateFunctions = updateNonBfFunctions;
			tuning = false;
			bestIsBf = false;
			break;

		case MODE_HYBRID:
			tuning = true;
			bfBetterCounter = 0;
			break;
		}

		executionsSinceStatsCollection = 0;
	}
	else if (tuning)
	{
		// First iteration is BF; then non-BF.
		bool executeBf = executionsSinceStatsCollection % 2;
		findFunctions = executeBf ? findBfFunctions : findNonBfFunctions;
		updateFunctions = executeBf ? updateBfFunctions : updateNonBfFunctions;

		LARGE_INTEGER startTime, endTime;
		QueryPerformanceCounter(&startTime);
		horizon = ScheduleGenerator(solution, times, executeBf);
		QueryPerformanceCounter(&endTime);
		int time = endTime.QuadPart - startTime.QuadPart;

		if (executeBf)
			bfTiming = time;
		else
		{
			// In this branch time is the timing of non-bf implementation

			if (bfTiming < time)
				bfBetterCounter++;

			int nonBfBetterCounter = executionsSinceStatsCollection / 2 - bfBetterCounter;
			if (bfBetterCounter < nonBfBetterCounter) // If non-BF is better in more than half the executions
			{
				if (bfBetterCounter <= GetMaxCounterForConfidence(executionsSinceStatsCollection))
				{
					bestIsBf = false;
					tuning = false;
					//printf("Decided Non-BF (%d) after %d iterations\n", nonBfBetterCounter, executionsSinceStatsCollection);
					// No need to update functions as this is a non-BF iteration
				}
			}
			else
			{
				//printf("BF is better than non-BF in %d pc cases.  BF timing: %d, non-BF timing: %d\n", bfBetterCounter * 200 / executionsSinceStatsCollection, bfTiming, time);
				if (nonBfBetterCounter <= GetMaxCounterForConfidence(executionsSinceStatsCollection))
				{
					bestIsBf = true;
					tuning = false;
					findFunctions = findBfFunctions;
					updateFunctions = updateBfFunctions;
					//printf("Decided BF (%d) after %d iterations\n", bfBetterCounter, executionsSinceStatsCollection);
				}
			} 
		}
	}
	else
		horizon = ScheduleGenerator(solution, times, bestIsBf);

	return horizon;
}


bool IsFeasibleSequence(int *solution)
{
	bool *scheduled = new bool[numberOfJobs];
	memset(scheduled, 0, numberOfJobs * sizeof(bool));

	for (int i = 0; i < numberOfJobs; i++)
	{
		int job = solution[i];

		if (scheduled[job])
			return false;

		for (int *predecessor = jobInfos[job].predecessors; *predecessor >= 0; predecessor++)
			if (!scheduled[*predecessor])
				return false;

		scheduled[job] = true;
	}

	delete[] scheduled;

	return true;
}

int ConventionalScheduleGenerator(int *solution, int *times)
{
	return ScheduleGenerator(solution, times, false);
}

__declspec(dllexport) int Metaheuristic(int *curSolution, int *bestSolution, int *times, int iterations, int randomSeed, int *runtimes)
{
	int (*generator)(int*, int*);
	if (mode == MODE_CONVENTIONAL)
	{
		findFunctions = findNonBfFunctions;
		updateFunctions = updateNonBfFunctions;
		generator = ConventionalScheduleGenerator;
	}
	else
		generator = ScheduleGenerator;

	srand(randomSeed);

	slotsTested = restarts = 0;

	int bestObjective = INT_MAX;
	int curObjective = generator(curSolution, times);

	int *curSolutionCopy = new int[numberOfJobs];
	memcpy(curSolutionCopy, curSolution, numberOfJobs * sizeof(int));

	int *predecessorsMap = new int[numberOfJobs];
	int *successorsMap = new int[numberOfJobs];
	memset(predecessorsMap, 0, numberOfJobs * sizeof(int));
	memset(successorsMap, 0, numberOfJobs * sizeof(int));

	_ASSERT(IsFeasibleSequence(curSolutionCopy));


	for (int iteration = 0; iteration < iterations; iteration++)
	{
		// Make a move
		int oldPos = rand() % numberOfJobs;
		int job = curSolution[oldPos];
		JobInfo *jobInfo = jobInfos + job;

		// Produce maps to then quickly scan for the range of feasible positions for the selected job
		for (int *predecessor = jobInfo->predecessors; *predecessor >= 0; predecessor++)
			predecessorsMap[*predecessor] = 1;
		for (int *successor = jobInfo->successors; *successor >= 0; successor++)
			successorsMap[*successor] = 1;

		// Find the range of feasible positions for the selected job
		int earliestPos = oldPos;
		while (--earliestPos >= 0)
			if (predecessorsMap[curSolution[earliestPos]])
				break;
		earliestPos++;

		int latestPos = oldPos;
		while (++latestPos < numberOfJobs)
			if (successorsMap[curSolution[latestPos]])
				break;
		latestPos--;



		// If the job is locked
		if (latestPos == earliestPos)
			continue;


		// Generate a feasible new position
		int newPos;
		do
		{
			newPos = rand() % (latestPos - earliestPos + 1) + earliestPos;
		} while (newPos == oldPos);


		// Apply the move
		if (newPos > oldPos)
		{
			for (int i = oldPos; i < newPos; i++)
				curSolutionCopy[i] = curSolutionCopy[i + 1];
			curSolutionCopy[newPos] = job;
		}
		else
		{
			for (int i = oldPos; i > newPos; i--)
				curSolutionCopy[i] = curSolutionCopy[i - 1];
			curSolutionCopy[newPos] = job;
		}

		_ASSERT(IsFeasibleSequence(curSolutionCopy));

		// Evaluate newSolution and decide whether to accept the move 
		// (accept all non-worsening; accept worsening with 50% probability)
		LARGE_INTEGER startTime, endTime;

#ifdef STORE_RUNTIMES
			QueryPerformanceCounter(&startTime);
#endif

		int newObjective = generator(curSolutionCopy, times);

#ifdef STORE_RUNTIMES
		QueryPerformanceCounter(&endTime);
		runtimes[iteration] = endTime.QuadPart - startTime.QuadPart;
#endif

		bool accept = newObjective <= curObjective || rand() % 2 < 1;

		// Save best found so far
		if (newObjective < bestObjective)
		{
			bestObjective = newObjective;
			memcpy(bestSolution, curSolutionCopy, numberOfJobs * sizeof(int));
			_ASSERT(generator(bestSolution, times) == bestObjective);
		}
		
		// Accept or reject the move
		if (accept)
		{
			curObjective = newObjective;
			memcpy(curSolution, curSolutionCopy, numberOfJobs * sizeof(int));
		}
		else
			memcpy(curSolutionCopy, curSolution, numberOfJobs * sizeof(int));


		//if (!(iteration % 100000))
		//	printf("Best %d\n", bestObjective);

		// Reset the maps
		for (int *predecessor = jobInfo->predecessors; *predecessor >= 0; predecessor++)
			predecessorsMap[*predecessor] = 0;
		for (int *successor = jobInfo->successors; *successor >= 0; successor++)
			successorsMap[*successor] = 0;
	}

	//printf("Per job: %.2f restarts, %.2f slots tested\n", restarts * 1.0 / iterations / numberOfJobs, slotsTested * 1.0 / iterations / numberOfJobs);

	delete[] predecessorsMap;
	delete[] successorsMap;
	delete[] curSolutionCopy;

	IsFeasibleSequence(bestSolution);

	int finalObjective = generator(bestSolution, times);
	_ASSERT(finalObjective == bestObjective);
	return finalObjective;
}


