#pragma once

#include "stdafx.h"

struct JobInfo;

typedef int(*find)(JobInfo *jobInfo, int startTime);
typedef void(*update)(JobInfo *jobInfo, int startTime);

struct JobInfo
{
	int *predecessors;
	int *successors;
	int duration;
	int *consumedResources;
	int *consumedVolumes;
	//int (*findEarliestSlotCollectStats)(JobInfo *jobInfo, int startTime);
	//void (*updateResourceAvailability)(JobInfo *jobInfo, int startTime);
	int bloomFilter;
	int unconsumedResourcesBfMask;
	int *resourcesToTest;
	int *volumesToTest;
	int *failedTests;
};


struct BloomFilterBit
{
	int resource;
	int minVolume;
	int demandProbability;
	int availabilityProbability;
	int nextIndex;
};

extern int numberOfJobs;
extern int numberOfResources;
extern int *resourceCapacities;
extern int **resourceAvailability;

// Allocated memory to store information within JobInfos
extern int *predecessorsFlat;
extern int *successorsFlat;
extern int *consumptionFlat;
extern int *consumedResourcesFlat;
extern int *consumedCapacitiesFlat;

extern JobInfo *jobInfos;

// Function sets used by SSGS
extern find *findFunctions;
extern update *updateFunctions;

// Function sets with various implementations, to be assigned to findFunctions and updateFunctions
extern find *findBfFunctions;
extern find *findNonBfFunctions;
extern find *findStatFunctions;
extern update *updateBfFunctions;
extern update *updateNonBfFunctions;

// Information related to Bloom filters 
extern int *bloomFilters;
extern int **bfDictionary;
extern int bfFullAvailability;
extern int *bfResourcesFlat;
extern int *bfCapacitiesFlat;
extern BloomFilterBit *bfBits;

extern int falsePositives;


int GetCorrectBf(int slot);
