#include "StatisticsCollection.h"
#include <crtdbg.h>
#include "AlgorithmParameters.h"

int **availabilityStat;
int *failedTestsFlat;
int slotsTested, restarts;

void InitStatistics()
{
	availabilityStat = new int*[numberOfResources];
	for (int resource = 0; resource < numberOfResources; resource++)
		availabilityStat[resource] = new int[resourceCapacities[resource] + 1];

	failedTestsFlat = new int[numberOfResources * numberOfJobs];

	int *curFailedTestsFlat = failedTestsFlat;
	for (JobInfo *jobInfo = jobInfos + numberOfJobs - 1; jobInfo >= jobInfos; jobInfo--)
	{
		jobInfo->failedTests = curFailedTestsFlat;
		curFailedTestsFlat += numberOfResources;
	}

	ResetStatistics();
}

void ClearStatistics()
{
	for (int resource = 0; resource < numberOfResources; resource++)
		delete[] availabilityStat[resource];

	delete[] availabilityStat;

	delete[] failedTestsFlat;

#ifdef _DEBUG
	availabilityStat = nullptr;
	failedTestsFlat = nullptr;
#endif
}

void ResetStatistics()
{
	for (int resource = 0; resource < numberOfResources; resource++)
		memset(availabilityStat[resource], 0, (resourceCapacities[resource] + 1) * sizeof(int));

	for (JobInfo *jobInfo = jobInfos + numberOfJobs - 1; jobInfo >= jobInfos; jobInfo--)
		memset(jobInfo->failedTests, 0, numberOfResources * sizeof(int));
}

int FindEarliestSlotCollectStats(JobInfo *jobInfo, int startTime)
{
	int earliestSlotToCheck = startTime;
	while (true)
	{
		int endTime = startTime + jobInfo->duration - 1;

		for (int slot = endTime; slot >= earliestSlotToCheck; slot--)
		{
			bool success = true;
			for (int *resource = jobInfo->consumedResources, *volume = jobInfo->consumedVolumes; *resource >= 0; resource++, volume++)
			{
				int curAvail = resourceAvailability[*resource][slot];
				availabilityStat[*resource][curAvail]++;
				if (curAvail < *volume)
				{
					success = false;
					jobInfo->failedTests[*resource]++;
				}
			}

			if (!success)
			{
				startTime = slot + 1;
				earliestSlotToCheck = endTime + 1;
				goto continueSearch;
			}
		}


		return startTime;

	continueSearch:;
	}
}

void Swap(int *x1, int *x2)
{
	int temp = *x1;
	*x1 = *x2;
	*x2 = temp;
}

void OrderResources(int *resources, int *volumes, int *stats)
{
	int n = 0;
	while (resources[n] >= 0)
		n++;

	for (int i = 0; i < n - 1; i++)
	{
		bool modified = false;
		for (int j = 0; j < n - i; j++)
		{
			if (stats[resources[j]] < stats[resources[j + 1]])
			{
				Swap(resources + j, resources + j + 1);
				Swap(volumes + j, volumes + j + 1);
				modified = true;
			}
		}
		
		if (!modified)
			break;
	}
}

void OptimiseResourcesOrder(JobInfo *jobInfo)
{
	return;
	OrderResources(jobInfo->consumedResources, jobInfo->consumedVolumes, jobInfo->failedTests);
	OrderResources(jobInfo->resourcesToTest, jobInfo->volumesToTest, jobInfo->failedTests);

#ifdef _DEBUG
	if (*jobInfo->consumedResources >= 0)
		for (int *resource = jobInfo->consumedResources + 1; *resource >= 0; resource++)
			_ASSERT(jobInfo->failedTests[*(resource - 1)] >= jobInfo->failedTests[*resource]);

	if (*jobInfo->resourcesToTest >= 0)
		for (int *resource = jobInfo->resourcesToTest + 1; *resource >= 0; resource++)
			_ASSERT(jobInfo->failedTests[*(resource - 1)] >= jobInfo->failedTests[*resource]);
#endif
}

void OptimiseResourcesOrder()
{
	for (JobInfo *jobInfo = jobInfos + numberOfJobs - 1; jobInfo >= jobInfos; jobInfo--)
		OptimiseResourcesOrder(jobInfo);
}