#pragma once
#include "stdafx.h"

#include <crtdbg.h>
#include "InstanceData.h"


int FindEarliestSlotTestBfOnly(JobInfo *jobInfo, int startTime)
{
	int earliestSlotToCheck = startTime;
	while (true)
	{
		int endTime = startTime + jobInfo->duration - 1;

		for (int slot = endTime; slot >= earliestSlotToCheck; slot--)
			if ((bloomFilters[slot] & jobInfo->bloomFilter) != jobInfo->bloomFilter)
			{
				startTime = slot + 1;
				earliestSlotToCheck = endTime + 1;
				goto continueSearch;
			}

		return startTime;

	continueSearch:;
	}
}


int FindEarliestSlotTestBfAndSingleResource(JobInfo *jobInfo, int startTime)
{
	int resource = *jobInfo->resourcesToTest;
	int volume = *jobInfo->volumesToTest;
	int* curAvailability = resourceAvailability[resource];

	int earliestSlotToCheck = startTime;
	while (true)
	{
		int endTime = startTime + jobInfo->duration - 1;

		// Test Bloom filters only
		for (int slot = endTime; slot >= earliestSlotToCheck; slot--)
			if ((bloomFilters[slot] & jobInfo->bloomFilter) != jobInfo->bloomFilter)
			{
				startTime = slot + 1;
				earliestSlotToCheck = slot + 1; // Need to test everything as accuracy is not guaranteed
				goto continueSearch;
			}

		// Test resources accurately
		for (int slot = endTime; slot >= earliestSlotToCheck; slot--)
			if (resourceAvailability[resource][slot] < volume)
			{
				startTime = slot + 1;
				earliestSlotToCheck = endTime + 1;
				falsePositives++;
				goto continueSearch;
			}

		return startTime;

	continueSearch:;
	}
}

int FindEarliestSlotTestBfAndSomeResources(JobInfo *jobInfo, int startTime)
{
	int earliestSlotToCheck = startTime;
	while (true)
	{
		int endTime = startTime + jobInfo->duration - 1;

		// Test Bloom filters only
		for (int slot = endTime; slot >= earliestSlotToCheck; slot--)
			if ((bloomFilters[slot] & jobInfo->bloomFilter) != jobInfo->bloomFilter)
			{
				startTime = slot + 1;
				earliestSlotToCheck = slot + 1; // Need to test everything as accuracy is not guaranteed
				goto continueSearch;
			}

		// Test resources accurately
		for (int slot = endTime; slot >= earliestSlotToCheck; slot--)
			for (int *resource = jobInfo->resourcesToTest, *volume = jobInfo->volumesToTest; *resource >= 0; resource++, volume++)
				if (resourceAvailability[*resource][slot] < *volume)
				{
					startTime = slot + 1;
					earliestSlotToCheck = endTime + 1;
					falsePositives++;
					goto continueSearch;
				}

		return startTime;

	continueSearch:;
	}
}


void UpdateResourseAvailabilityBFSingle(JobInfo *jobInfo, int startTime)
{
	int *resource = jobInfo->consumedResources;
	for (int slot = startTime + jobInfo->duration - 1; slot >= startTime; slot--)
	{
		int newAvailability = resourceAvailability[*resource][slot] - *jobInfo->consumedVolumes;
		resourceAvailability[*resource][slot] = newAvailability;

		bloomFilters[slot] = (bloomFilters[slot] & jobInfo->unconsumedResourcesBfMask)
			| bfDictionary[*resource][newAvailability];
	}
}


void UpdateResourceAvailabilityBf(JobInfo *jobInfo, int startTime)
{
	for (int slot = startTime + jobInfo->duration - 1; slot >= startTime; slot--)
	{
		_ASSERT(GetCorrectBf(slot) == bloomFilters[slot]);

		int filter = bloomFilters[slot] & jobInfo->unconsumedResourcesBfMask;

		for (int *resource = jobInfo->consumedResources, *volume = jobInfo->consumedVolumes; *resource >= 0; resource++, volume++)
		{
			int newAvailability = resourceAvailability[*resource][slot] - *volume;
			resourceAvailability[*resource][slot] = newAvailability;

			filter |= bfDictionary[*resource][newAvailability];
		}

		_ASSERT(GetCorrectBf(slot) == filter);

		bloomFilters[slot] = filter;
	}
}


